/*
 * yasapi_config_control.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static void CheckBoxTypeInit(const Control *pControl, HWND hDlg)
{
}

static void CheckBoxTypeSet(const Control *pControl, HWND hDlg,
    const void *pData)
{
  const ControlCheckBoxConfig *config=pControl->config;
  int val=VALUE_INT(pData,pControl->offset);

  CheckDlgButton(
    hDlg,                 // _In_  HWND hDlg,
    config->idc,          // _In_  int nIDButton,
    val==config->on       // _In_  UINT uCheck
  );
}

static void CheckBoxTypeGet(const Control *pControl, HWND hDlg,
    void *pData)
{
  const ControlCheckBoxConfig *config=pControl->config;
  UINT uChecked;

  uChecked=IsDlgButtonChecked(
    hDlg,                 // _In_  HWND hDlg,
    config->idc           // _In_  int nIDButton
  );

  VALUE_INT(pData,pControl->offset)=uChecked?config->on:config->off;
}

const ControlType gcCheckBoxType={
  CheckBoxTypeInit,
  CheckBoxTypeSet,
  CheckBoxTypeGet
};

///////////////////////////////////////////////////////////////////////////////
static void RadioButtonTypeInit(const Control *pControl, HWND hDlg)
{
}

static void RadioButtonTypeSet(const Control *pControl, HWND hDlg,
    const void *pData)
{
  const ControlRadioButtonConfig *config=pControl->config;
  int val=VALUE_INT(pData,pControl->offset);

  while (config->idc) {
    CheckDlgButton(
      hDlg,             // _In_  HWND hDlg,
      config->idc,      // _In_  int nIDButton,
      val==config->val  // _In_  UINT uCheck
    );

    ++config;
  }
}

static void RadioButtonTypeGet(const Control *pControl, HWND hDlg,
    void *pData)
{
  const ControlRadioButtonConfig *config=pControl->config;
  UINT uChecked;

  while (config->idc) {
    uChecked=IsDlgButtonChecked(
      hDlg,             // _In_  HWND hDlg,
      config->idc       // _In_  int nIDButton
    );

    if (uChecked) {
      VALUE_INT(pData,pControl->offset)=config->val;
      break;
    }

    ++config;
  }
}

const ControlType gcRadioButtonType={
  RadioButtonTypeInit,
  RadioButtonTypeSet,
  RadioButtonTypeGet
};

///////////////////////////////////////////////////////////////////////////////
static void SliderTypeInit(const Control *pControl, HWND hDlg)
{
}

static void SliderTypeSet(const Control *pControl, HWND hDlg,
    const void *pData)
{
  double val=VALUE_DOUBLE(pData,pControl->offset);
  const ControlSliderConfig *config=pControl->config;
  HWND hWndCtl=GetDlgItem(hDlg,config->idc);
  LRESULT lMin=SendMessageW(hWndCtl,TBM_GETRANGEMIN,0,0);
  LRESULT lMax=SendMessageW(hWndCtl,TBM_GETRANGEMAX,0,0);
  LPARAM lPos=(double)lMin
      +(val-config->min)/(config->max-config->min)*(lMax-lMin)
      +0.5;

  SendMessageW(hWndCtl,TBM_SETPOS,TRUE,lPos);
}

static void SliderTypeGet(const Control *pControl, HWND hDlg,
    void *pData)
{
  const ControlSliderConfig *config=pControl->config;
  HWND hWndCtl=GetDlgItem(hDlg,config->idc);
  LRESULT lMin=SendMessageW(hWndCtl,TBM_GETRANGEMIN,0,0);
  LRESULT lMax=SendMessageW(hWndCtl,TBM_GETRANGEMAX,0,0);
  LRESULT lPos=SendMessageW(hWndCtl,TBM_GETPOS,0,0);

  VALUE_DOUBLE(pData,pControl->offset)=config->min
      +((double)lPos-lMin)/(lMax-lMin)*(config->max-config->min);
}

const ControlType gcSliderType={
  SliderTypeInit,
  SliderTypeSet,
  SliderTypeGet
};

///////////////////////////////////////////////////////////////////////////////
static void ComboBoxTypeInit(const Control *pControl, HWND hDlg)
{
  const ControlComboBoxConfig *config=pControl->config;
  const ControlComboBoxList *list=config->list;
  HWND hWndCtl=GetDlgItem(hDlg,config->idc);

  while (list->label) {
    SendMessageW(hWndCtl,CB_ADDSTRING,0,(LPARAM)list->label);
    ++list;
  }
}

static void ComboBoxTypeSet(const Control *pControl, HWND hDlg,
    const void *pData)
{
  int val=VALUE_INT(pData,pControl->offset);
  const ControlComboBoxConfig *config=pControl->config;
  const ControlComboBoxList *list=config->list;
  HWND hWndCtl=GetDlgItem(hDlg,config->idc);
  WPARAM wParam=0;

  while (list->label) {
    if (val==list->val) {
      SendMessageW(hWndCtl,CB_SETCURSEL,wParam,0);
      break;
    }

    ++wParam;
    ++list;
  }
}

static void ComboBoxTypeGet(const Control *pControl, HWND hDlg,
    void *pData)
{
  const ControlComboBoxConfig *config=pControl->config;
  const ControlComboBoxList *list=config->list;
  HWND hWndCtl=GetDlgItem(hDlg,config->idc);
  LRESULT lr=SendMessageW(hWndCtl,CB_GETCURSEL,0,0);

  VALUE_INT(pData,pControl->offset)=list[lr].val;
}

const ControlType gcComboBoxType={
  ComboBoxTypeInit,
  ComboBoxTypeSet,
  ComboBoxTypeGet
};

///////////////////////////////////////////////////////////////////////////////
void ControlInit(const Control *pControl, HWND hDlg)
{
  while (pControl->type) {
    pControl->type->Init(pControl,hDlg);
    ++pControl;
  }
}

void ControlSet(const Control *pControl, HWND hDlg, const void *pData)
{
  while (pControl->type) {
    pControl->type->Set(pControl,hDlg,pData);
    ++pControl;
  }
}

void ControlGet(const Control *pControl, HWND hDlg, void *pData)
{
  while (pControl->type) {
    pControl->type->Get(pControl,hDlg,pData);
    ++pControl;
  }
}
