/*
 * yasapi_session.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
// Rendering a Stream
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd316756%28v=vs.85%29.aspx

// Exclusive-Mode Streams
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd370844%28v=vs.85%29.aspx

#if ! defined (AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY) // {
#define AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY 0x08000000
#endif // }

#if ! defined(AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM) // {
#define AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM 0x80000000
#endif // }

///////////////////////////////////////////////////////////////////////////////
int SessionCreate(Session *pSession, const SessionCreateConfig *pConfig,
    Channel *pChannel)
{
  enum { WINAMP_PACKET_SIZE=576 };
  Plugin *pPlugin=pChannel->pPlayer->pPlugin;
  int debug=pPlugin->options.debug;
  int samplerate=pConfig->samplerate;
  int numchannels=pConfig->numchannels;
  int bitspersamp=pConfig->bitspersamp;
  WAVEFORMATEXTENSIBLE *pwfxx=&pSession->wfxx;
  WAVEFORMATEX *pwfx=&pwfxx->Format;
  IMMDevice *pDevice=pChannel->pDevice;
  double qShareSize=pChannel->options.qShareSize;
  double qRingFill=pChannel->options.qRingFill;
  double qRingSize=pChannel->options.qRingSize;
  REFERENCE_TIME hnsDevicePeriod;
  int retry;
  AUDCLNT_SHAREMODE eShareMode;
  DWORD dwStreamFlags;
  IAudioClient *pClient;
  IAudioClock *pClock;
  IAudioRenderClient *pRender;
  UINT32 uBufSize,uBufDelta;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  pSession->pChannel=pChannel;
  pSession->pStrategy=pChannel->options.bPull?&gcsPull:&gcsPush;

  // clear state //////////////////////////////////////////////////////////////
  pSession->state.valid=0;
  pSession->state.started=0;
  pSession->state.pause=0;
  pSession->state.wait=0;
  pSession->state.drain=0;
  pSession->state.empty=0;

  // setup session parameter //////////////////////////////////////////////////
  DPRINTF(debug,0,"  samplerate: %d\n",samplerate);

  if (1==numchannels&&pPlugin->options.mono2stereo) {
    DPUTS(debug,0,"  numchannels: 1 -> 2\n");
    pSession->bMono2Stereo=TRUE;
    numchannels=2;
  }
  else {
    DPRINTF(debug,0,"  numchannels: %d\n",numchannels);
    pSession->bMono2Stereo=FALSE;
  }

  DPRINTF(debug,0,"  bitspersamp: %d\n",bitspersamp);
  WFXXSetup(pwfxx,samplerate,numchannels,bitspersamp,FALSE,FALSE);
  pSession->wBytesPerSample=pwfx->wBitsPerSample>>3;

  // setup share mode /////////////////////////////////////////////////////////
  switch (pChannel->options.eShareMode) {
  case YASAPI_SHAREMODE_EXCLUSIVE:
  case YASAPI_SHAREMODE_AUTOMATIC:
    eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
    break;
  default:
    eShareMode=AUDCLNT_SHAREMODE_SHARED;
    break;
  }

  /////////////////////////////////////////////////////////////////////////////
  hnsDevicePeriod=0;
  retry=0;
retry:
  // activate the audio client ////////////////////////////////////////////////
  hr=pDevice->lpVtbl->Activate(pDevice,
    &IID_IAudioClient,    // [in]   REFIID iid,
    CLSCTX_ALL,           // [in]   DWORD dwClsCtx,
    NULL,                 // [in]   PROPVARIANT *pActivationParams,
    (void **)&pClient     // [out]  void **ppInterface
  );

  if (FAILED(hr)) {
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"activating audio client");
    goto client;
  }

  pSession->pClient=pClient;
  DPUTS(debug,0,"  audio client created\n");

  // setup the stream flags ///////////////////////////////////////////////////
  dwStreamFlags=0;

  if (AUDCLNT_SHAREMODE_SHARED==eShareMode
        &&pChannel->options.bAutoConvertPCM) {
      dwStreamFlags|=AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM;

     if (pChannel->options.bSRCDefaultQuality)
       dwStreamFlags|=AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;
  }

  // if needed, get the minimum device period (low latency) ///////////////////
  // we may have it already because of not aligned buffer error, cf. below.
  if (0==hnsDevicePeriod) {
#if 0 // {
    hr=pClient->lpVtbl->GetDevicePeriod(pClient,
      &hnsDevicePeriod,   // [out]  REFERENCE_TIME *phnsDefaultDevicePeriod,
      NULL                // [out]  REFERENCE_TIME *phnsMinimumDevicePeriod
    );
#else // } {
    hr=pClient->lpVtbl->GetDevicePeriod(pClient,
      NULL,               // [out]  REFERENCE_TIME *phnsDefaultDevicePeriod,
      &hnsDevicePeriod    // [out]  REFERENCE_TIME *phnsMinimumDevicePeriod
    );
#endif // }

    if (FAILED(hr)) {
      DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DERROR(debug,E_POINTER,hr);
      DUNKNOWN(debug,hr);
      DMESSAGE(debug,"getting device period");
      goto period;
    }

    DPRINTF(debug,0,"  got device period (%I64d)\n",hnsDevicePeriod);

    if (1.0<qShareSize) {
      hnsDevicePeriod=qShareSize*hnsDevicePeriod+0.5;
      DPRINTF(debug,0,"  device period scaled by %.02f (device period: %d)\n",
          qShareSize,hnsDevicePeriod);
    }
  }

  // initialize the audio client //////////////////////////////////////////////
  hr=pSession->pStrategy->Initialize(pSession,eShareMode,dwStreamFlags,
      hnsDevicePeriod);

  if (FAILED(hr)) {
    if (AUDCLNT_SHAREMODE_EXCLUSIVE==eShareMode) {
      switch (hr) {
      case AUDCLNT_E_ENDPOINT_CREATE_FAILED:
        if (YASAPI_SHAREMODE_AUTOMATIC!=pChannel->options.eShareMode)
          break;

        DPUTS(debug,0,"  failed with AUDCLNT_E_ENDPOINT_CREATE_FAILED\n");
        DPUTS(debug,0,"  destroying audio client\n");
        pClient->lpVtbl->Release(pClient);

        DPUTS(debug,0,"  retrying in shared mode\n");
        eShareMode=AUDCLNT_SHAREMODE_SHARED;
        goto retry;
      case AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED:
        if (0<retry)
          break;

        DPUTS(debug,0,"  failed with AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED\n");
        DPUTS(debug,0,"  trying to get an aligned buffer size\n");
        retry=1;

        // 1. Call IAudioClient::GetBufferSize and receive the
        //    next-highest-aligned buffer size (in frames).
        hr=pClient->lpVtbl->GetBufferSize(pClient,
          &pSession->uFrames    // [out] UINT32 *pNumBufferFrames
        );

        if (FAILED(hr)) {
          DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
          DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
          DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
          DERROR(debug,E_POINTER,hr);
          DUNKNOWN(debug,hr);
          DMESSAGE(debug,"getting aligned number of frames");
          goto init;
        }

        DPRINTF(debug,0,"  got aligned number of frames: %u\n",
            pSession->uFrames);

        // 3. Calculate the aligned buffer size in 100-nansecond units (hns).
        //    The buffer size is (REFERENCE_TIME)((10000.0 * 1000 /
        //    WAVEFORMATEX.nSamplesPerSecond * nFrames) + 0.5). In this
        //    formula, nFrames is the buffer size retrieved by GetBufferSize.
        hnsDevicePeriod=YASAPI_FRAMES_TIMER(pSession->uFrames,pwfx)+0.5;

#if 0 // {
        // save hnsDevicePeriod into device options.
        OptionsSetDevicePeriod(pPlugin->options.hnsDevicePeriod);
#endif // }

        // 2. Call IAudioClient::Release to release the audio
        //    client used in the previous call that returned
        //    AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED.
        DPUTS(debug,0,"  destroying audio client\n");
        pClient->lpVtbl->Release(pClient);

        // 4. Call the IMMDevice::Activate method with parameter iid set
        //    to REFIID IID_IAudioClient to create a new audio client.
        // 5. Call Initialize again on the created audio client and
        //    specify the new buffer size and periodicity.
        DPUTS(debug,0,"  retrying with aligned buffer\n");
        goto retry;
      default:
        break;
      }
    }

    DERROR(debug,AUDCLNT_E_ALREADY_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_CPUUSAGE_EXCEEDED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_IN_USE,hr);
    DERROR(debug,AUDCLNT_E_ENDPOINT_CREATE_FAILED,hr);
    DERROR(debug,AUDCLNT_E_INVALID_DEVICE_PERIOD,hr);
    DERROR(debug,AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
    DERROR(debug,AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED,hr);
    DERROR(debug,AUDCLNT_E_BUFDURATION_PERIOD_NOT_EQUAL,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"initializing audio client");
    goto init;
  }

  DPUTS(debug,0,"  audio client initialized\n");

  // remember the share mode //////////////////////////////////////////////////
  pSession->eShareMode=eShareMode;

  // set the event handle /////////////////////////////////////////////////////
  if (pChannel->options.bPull) {
    hr=pClient->lpVtbl->SetEventHandle(pClient,
      pPlugin->hEvent       // [in] HANDLE eventHandle
    );

    if (FAILED(hr)) {
      DERROR(debug,E_INVALIDARG,hr);
      DERROR(debug,AUDCLNT_E_EVENTHANDLE_NOT_EXPECTED,hr);
      DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
      DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DUNKNOWN(debug,hr);
      DMESSAGE(debug,"setting event handle");
      goto clock;
    }

    DPUTS(debug,0,"  event handle set\n");
  }

  // get the audio clock //////////////////////////////////////////////////////
  hr=pClient->lpVtbl->GetService(pClient,
    &IID_IAudioClock,       // [in]   REFIID riid,
    (void**)&pClock         // [out]  void **ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"getting audio clock");
    goto clock;
  }

  pSession->pClock=pClock;
  DPUTS(debug,0,"  audio clock created\n");

  // get the buffer size //////////////////////////////////////////////////////
  //
  // This method retrieves the length of the endpoint buffer shared
  // between the client application and the audio engine. The length
  // is expressed as the number of audio frames the buffer can hold.
  // The size in bytes of an audio frame is calculated as the number
  // of channels in the stream multiplied by the sample size per channel.
  // For example, the frame size is four bytes for a stereo (2-channel)
  // stream with 16-bit samples.
  hr=pClient->lpVtbl->GetBufferSize(pClient,
    &pSession->uFrames        // [out]  UINT32 *pNumBufferFrames
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"getting buffer size");
    goto size;
  }

  DPRINTF(debug,0,"  got buffer size (frames: %d)\n",pSession->uFrames);
  pSession->uFramesMin=
      1.0==qRingFill?pSession->uFrames:
      (UINT32)(qRingFill*pSession->uFrames+0.5);
  pSession->uFramesPadding=0;
  pSession->time=pSession->pStrategy->GetTime(pSession,eShareMode);

  // get the render client ////////////////////////////////////////////////////
  hr=pClient->lpVtbl->GetService(pClient,
    &IID_IAudioRenderClient,  // [in]   REFIID riid,
    (void**)&pRender          // [out]  void **ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"getting audio render client");
    goto render;
  }

  pSession->pRender=pRender;
  DPUTS(debug,0,"  audio render client created\n");

  // create the ring buffer ///////////////////////////////////////////////////
  uBufSize=qRingSize*(pSession->uFrames/qShareSize)+0.5;

  // we should at least hold the minimum requiered frames.
  if (uBufSize<pSession->uFramesMin)
    uBufSize=pSession->uFramesMin;

  // round up to a multiple of "WINAMP_PACKET_SIZE".
  if (0<(uBufDelta=uBufSize%WINAMP_PACKET_SIZE))
    uBufSize+=WINAMP_PACKET_SIZE-uBufDelta;

  DPRINTF(debug,0,"  ring buffer size: %0.2f\n",
      qShareSize*uBufSize/pSession->uFrames);

  // "nBlockAlign" reflects "bMono2Stereo".
  uBufSize*=pwfx->nBlockAlign;

  if (RingCreate(&pSession->ring,uBufSize,debug)<0) {
    DMESSAGE(debug,"creating ring buffer\n");
    goto ring;
  }

  pSession->ring.Copy=SessionCopyMemory;
  pSession->ring.pClient=pSession;
  DPUTS(debug,0,"  ring buffer created\n");

  // success //////////////////////////////////////////////////////////////////
  pSession->state.valid=1;

  return YASAPI_FRAMES_MS(pSession->uFramesMin,pwfx);
// cleanup:
  //RingDestroy(&pSession->ring);
ring:
  pRender->lpVtbl->Release(pRender);
render:
size:
  pClock->lpVtbl->Release(pClock);
clock:
init:
period:
  pClient->lpVtbl->Release(pClient);
client:
  memset(pSession,0,sizeof *pSession);
DPRINTF(debug,0,"%s: returning with error\n",__func__);
  return -1;
}

void SessionDestroy(Session *pSession)
{
#if defined (YASAPI_DEBUG) // {
  int debug=SessionGetDebug(pSession);
#endif // }
  IAudioRenderClient *pRender=pSession->pRender;
  IAudioClock *pClock=pSession->pClock;
  IAudioClient *pClient=pSession->pClient;

  pSession->state.valid=0;

  DPUTS(debug,0,"  destroying ring buffer\n");
  RingDestroy(&pSession->ring);

  DPUTS(debug,0,"  destroying audio render client\n");
  pRender->lpVtbl->Release(pRender);

  DPUTS(debug,0,"  destroying audio clock\n");
  pClock->lpVtbl->Release(pClock);

  DPUTS(debug,0,"  destroying audio client\n");
  pClient->lpVtbl->Release(pClient);

  memset(pSession,0,sizeof *pSession);
}

///////////////////////////////////////////////////////////////////////////////
Plugin *SessionGetPlugin(Session *pSession)
{
  return pSession->pChannel->pPlayer->pPlugin;
}

#if defined (YASAPI_DEBUG) // {
int SessionGetDebug(Session *pSession)
{
  return SessionGetPlugin(pSession)->options.debug;
}
#endif // }

int SessionCanWrite(Session *pSession)
{
  return pSession->state.pause?0:
      !pSession->bMono2Stereo?pSession->ring.dwAvailable:
      pSession->ring.dwAvailable>>1;
}

///////////////////////////////////////////////////////////////////////////////
int SessionStart(Session *pSession)
{
#if defined (YASAPI_DEBUG) // {
  int debug=SessionGetDebug(pSession);
#endif // }
  WAVEFORMATEX *pwfx=&pSession->wfxx.Format;
  UINT32 uFramesRing=pSession->ring.dwWritten/pwfx->nBlockAlign;
  UINT32 uFramesPadding=pSession->uFramesPadding;
  UINT32 uFramesWritten=uFramesRing+uFramesPadding;
  IAudioClient *pClient=pSession->pClient;
  HRESULT hr;

  if (!pSession->state.started&&pSession->uFramesMin<=uFramesWritten) {
    // start the audio client /////////////////////////////////////////////////
    hr=pClient->lpVtbl->Start(pClient);

    if (FAILED(hr)) {
      DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
      DERROR(debug,AUDCLNT_E_NOT_STOPPED,hr);
      DERROR(debug,AUDCLNT_E_EVENTHANDLE_NOT_SET,hr);
      DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DUNKNOWN(debug,hr);
      DMESSAGE(debug,"starting audio client");
      goto start;
    }
  
    DPUTS(debug,0,"  audio client started\n");
    pSession->state.started=1;

    // start the timer ////////////////////////////////////////////////////////
    pSession->pStrategy->SetTimer(pSession,pSession->time,FALSE);
  }

  return 0;
start:
  return -1;
}

void SessionStop(Session *pSession)
{
#if defined (YASAPI_DEBUG) // {
  int debug=SessionGetDebug(pSession);
#endif // }
  IAudioClient *pClient=pSession->pClient;

  if (SESSION_IS_STARTED(pSession)) {
    // stop the timer /////////////////////////////////////////////////////////
    pSession->pStrategy->SetTimer(pSession,0,FALSE);

    // stop the audio client //////////////////////////////////////////////////
    DPUTS(debug,0,"  stopping audio client\n");
    pSession->state.started=0;
    pClient->lpVtbl->Stop(pClient);
    pSession->uFramesPadding=SessionGetFramesPadding(pSession);
  }
}

UINT32 SessionGetFramesPadding(Session *pSession)
{
#if defined (YASAPI_DEBUG) // {
  int debug=SessionGetDebug(pSession);
#endif // }
  IAudioClient *pClient=pSession->pClient;
  UINT32 uFramesPadding;
  HRESULT hr;

  hr=pClient->lpVtbl->GetCurrentPadding(pClient,
    &uFramesPadding       // [out]  UINT32 *pNumPaddingFrames
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"getting padding");
    goto padding;
  }

  return uFramesPadding;
padding:
  return -1;
}

int SessionWrite(Session *pSession, const SessionWriteConfig *pConfig)
{
#if defined (YASAPI_DEBUG) // {
  int debug=SessionGetDebug(pSession);
#endif // }
  Plugin *pPlugin=SessionGetPlugin(pSession);
  const char *buf=pConfig->buf;
  int len=pConfig->len;
  Ring *pRing=&pSession->ring;
  Queue *pQueue=&pPlugin->queue;
  HRESULT hr;

  if (pRing->dwAvailable<len) {
    DMESSAGE(debug,"ring buffer to small");
    goto avail;
  }

  if (pSession->bMono2Stereo)
    RingWrite(pRing,buf,len<<1,2);
  else
    RingWrite(pRing,buf,len,1);

  if (SESSION_IS_STARTED(pSession)&&pSession->state.wait) {
    pSession->state.wait=0;
    QueueWrite(pQueue,REQUEST_READ,NULL);
  }
  else if (!SESSION_IS_STARTED(pSession)&&!pSession->state.empty) {
    if (SessionStart(pSession)<0)
      goto start;
  }

  return 0;
start:
avail:
  return 1;
}

int SessionRead(Session *pSession, BOOL bFlush)
{
#if defined (YASAPI_DEBUG) // {
  int debug=SessionGetDebug(pSession);
#endif // }
  IAudioClient *pClient=pSession->pClient;
  IAudioRenderClient *pRender=pSession->pRender;
  Ring *pRing=&pSession->ring;
  WAVEFORMATEX *pwfx=&pSession->wfxx.Format;
  DWORD dwFlags=0;
  UINT32 uFramesPadding;
  UINT32 uFramesWrite;
  UINT32 uFramesRing;
  BYTE *pData;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  // in "state.empty" we should never go here.
  if (pSession->state.pause||pSession->state.empty)
    goto success;

  /////////////////////////////////////////////////////////////////////////////
  pSession->state.wait=0;

  // get padding //////////////////////////////////////////////////////////////
  if ((uFramesPadding=pSession->pStrategy->GetFramesPadding(pSession))<0)
    goto padding;

  uFramesWrite=pSession->uFrames-uFramesPadding;
  uFramesRing=pRing->dwWritten/pwfx->nBlockAlign;

  if (0==uFramesRing&&!bFlush) {
    if (pSession->state.drain)
      bFlush=TRUE;
    else {
      pSession->state.wait=1;
      goto success;
    }
  }

  if (bFlush) {
    pSession->state.empty=1;
    dwFlags|=AUDCLNT_BUFFERFLAGS_SILENT;
  }
  else if (uFramesWrite>uFramesRing) {
    if (pSession->pStrategy->CanAdapt(pSession))
      uFramesWrite=uFramesRing;
    else {
      pSession->state.wait=1;
      goto success;
    }
  }

  // get the render buffer ////////////////////////////////////////////////////
  hr=pRender->lpVtbl->GetBuffer(pRender,
    uFramesWrite,       // [in]   UINT32 NumFramesRequested,
    &pData              // [out]  BYTE **ppData
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_BUFFER_ERROR,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_TOO_LARGE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"getting render buffer");
    goto get;
  };

  if (!bFlush) {
    // copy the data from the ring buffer into the audio buffer ///////////////
    RingRead(pRing,pData,uFramesWrite*pwfx->nBlockAlign);
  }

  // release the render buffer ////////////////////////////////////////////////
  hr=pRender->lpVtbl->ReleaseBuffer(pRender,
    uFramesWrite,     // [in]  UINT32 NumFramesWritten,
    dwFlags           // [in]  DWORD dwFlags
  );

  if (FAILED(hr)) {
    crash("%s (%d)\n"
        "uFrames: %u\n"
        "uFramesWrite: %u\n"
        "uFramesRing: %u\n"
        "bFlush: %d\n",
        basename(__FILE__),__LINE__,
        pSession->uFrames,
        uFramesWrite,
        uFramesRing,
        bFlush);
    DERROR(debug,AUDCLNT_E_INVALID_SIZE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"releasing render buffer");
    goto release;
  }

  // start timer //////////////////////////////////////////////////////////////
  // "bFlush" and "state.empty" are equivalent. if set, the timer ist fired
  // regardless of being in shared or exclusive mode, push or pull.
  pSession->pStrategy->SetTimer(pSession,pSession->time,bFlush);
success:
  return 0;
release:
get:
padding:
  return -1;
}

UINT64 SessionGetPosition(Session *pSession)
{
#if defined (YASAPI_DEBUG) // {
  int debug=SessionGetDebug(pSession);
#endif // }
  IAudioClock *pClock=pSession->pClock;
  UINT64 u64Position;
  HRESULT hr;

  // get position /////////////////////////////////////////////////////////////
  hr=pClock->lpVtbl->GetPosition(pClock,
    &u64Position,         // [out] UINT64 *pu64Position,
    NULL                  // [out] UINT64 *pu64QPCPosition
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"getting position");
    goto position;
  }

  return u64Position;
position:
  return 0;
}

UINT64 SessionGetFrequency(Session *pSession)
{
#if defined (YASAPI_DEBUG) // {
  int debug=SessionGetDebug(pSession);
#endif // }
  IAudioClock *pClock=pSession->pClock;
  UINT64 u64Frequency;
  HRESULT hr;

  // get frequency ////////////////////////////////////////////////////////////
  hr=pClock->lpVtbl->GetFrequency(pClock,
    &u64Frequency         // [out] UINT64 *pu64Frequency
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"getting frequency");
    goto frequency;
  }

  return u64Frequency;
frequency:
  return 0;
}

int SessionGetTime(Session *pSession)
{
  WAVEFORMATEX *pwfx=&pSession->wfxx.Format;
  Ring *pRing=&pSession->ring;
  UINT32 uFramesRing=pRing->dwWritten/pwfx->nBlockAlign;
  UINT32 uFramesPadding=SessionGetFramesPadding(pSession);
  UINT64 u64Position=SessionGetPosition(pSession);
  UINT64 u64Frequency=SessionGetFrequency(pSession);
  double time=0.0;

  time+=(double)uFramesRing/pwfx->nSamplesPerSec;
  time+=(double)uFramesPadding/pwfx->nSamplesPerSec;

  if (u64Frequency)
    time+=(double)u64Position/u64Frequency;

  return 1000.0*time+0.5;
}

///////////////////////////////////////////////////////////////////////////////
void SessionSendRead(void *data)
{
  Session *pSession=data;
  Queue *pQueue;
  UINT32 uFrames;
  WAVEFORMATEX *pwfx;
  Request *pRequest;

#if 0 // {
  if (pSession->state.valid) {
#endif
    pQueue=&SessionGetPlugin(pSession)->queue;
    uFrames=pSession->uFrames;
    pwfx=&pSession->wfxx.Format;

    pRequest=QueueLockWrite(pQueue,NULL,REQUEST_READ);
    QueueUnlockWrite(pQueue,NULL);
#if 0
  }
#endif // }
}

VOID CALLBACK SessionFireTimer(LPVOID lpArg, DWORD dwLow, DWORD dwHigh)
{
  SessionSendRead(lpArg);
}

void SessionSetTimer(Session *pSession, LONGLONG time)
{
  HANDLE hTimer=SessionGetPlugin(pSession)->hTimer;
  LARGE_INTEGER ui;

  // http://msdn.microsoft.com/en-us/library/windows/desktop/ms687012%28v=vs.85%29.aspx
  // Positive values indicate absolute time. 
  // Negative values indicate relative time.
  // in 100 nanosecond intervals
  ui.QuadPart=-time;

  SetWaitableTimer(
    hTimer,           // _In_      HANDLE hTimer,
    &ui,              // _In_      const LARGE_INTEGER *pDueTime,
    0,                // _In_      LONG lPeriod,
    time?SessionFireTimer:NULL,
                      // _In_opt_  PTIMERAPCROUTINE pfnCompletionRoutine,
    time?pSession:NULL,
                      // _In_opt_  LPVOID lpArgToCompletionRoutine,
    FALSE             // _In_      BOOL fResume
  );
}

///////////////////////////////////////////////////////////////////////////////
void SessionCopyMemory(PVOID pClient, PVOID Destination, const VOID *Source,
    SIZE_T Length)
{
  Session *pSession=pClient;
  Plugin *pPlugin=pSession->pChannel->pPlayer->pPlugin;
  PluginOptions *pOptions=&pPlugin->options;
  double qVolume=pPlugin->qVolume;
  char *wp,*mp;
  const char *rp;
  int m,n;
  double q;
  int32_t i32;

  wp=Destination;
  mp=wp+Length;
  rp=Source;
  n=pSession->wfxx.Format.wBitsPerSample;
  m=n>>3;
  q=qVolume*YASAPI_INT_MAX(n)/YASAPI_INT_MAX(32);

  n=32-n;

  if (0==pSession->bMono2Stereo) {
    if (0==pOptions->volume||1.0==qVolume) {
      CopyMemory(
        Destination,            // _In_  PVOID Destination,
        Source,                 // _In_  const VOID *Source,
        Length                  // _In_  SIZE_T Length
      );
    }
    else {
      while (wp<mp) {
        CopyMemory(
          &i32,                 // _In_  PVOID Destination,
          rp,                   // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        i32<<=n;
        i32=q*i32+0.5;

        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          &i32,                 // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;
        rp+=m;
      }
    }
  }
  else {
    while (wp<mp) {
      CopyMemory(
        wp,                     // _In_  PVOID Destination,
        rp,                     // _In_  const VOID *Source,
        m                       // _In_  SIZE_T Length
      );

      wp+=m;

      CopyMemory(
        wp,                     // _In_  PVOID Destination,
        rp,                     // _In_  const VOID *Source,
        m                       // _In_  SIZE_T Length
      );

      wp+=m;
      rp+=m;
    }
  }
}
